/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"

static struct mog_fd *notifywr_mfd;
static struct mog_notifywr *notifywr;
static time_t usage_file_updated_at;
static time_t usage_file_interval = 10;
struct mog_queue *mog_notify_queue;

void mog_notify_init(void)
{
	const char *interval = getenv("MOG_DISK_USAGE_INTERVAL");
	struct mog_fd *mfd;
	int self_pipe[2];

	if (pipe2(self_pipe, O_NONBLOCK | O_CLOEXEC) < 0)
		die("failed to init self-pipe: %s\n", strerror(errno));

	if (interval) {
		int i = atoi(interval);

		if (i > 0)
			usage_file_interval = (time_t)i;
	}

	assert(mog_notify_queue == NULL && "notify queue already initialized");
	mog_notify_queue = mog_queue_new();

	mfd = mog_fd_init(self_pipe[0], MOG_FD_TYPE_NOTIFYRD);
	mfd->as.notifyrd.queue = mog_notify_queue;
	mog_idleq_push(mog_notify_queue, mfd, MOG_QEV_RD);

	notifywr_mfd = mfd = mog_fd_init(self_pipe[1], MOG_FD_TYPE_NOTIFYWR);
	notifywr = &mfd->as.notifywr;
	notifywr->queue = mog_notify_queue;
	memset(notifywr->notes, 0, sizeof(notifywr->notes));
}

static void global_mkusage(void)
{
	mog_mkusage_all();
	usage_file_updated_at = time(NULL);
}

/* just drain the pipe */
static void notifyrd_queue_step(struct mog_fd *mfd)
{
	char buf[128];
	ssize_t r;
	struct mog_notifyrd *notifyrd = &mfd->as.notifyrd;

	do {
		r = read(mfd->fd, buf, sizeof(buf));
	} while (r > 0 || (r < 0 && errno == EINTR));
	assert(r < 0 && errno == EAGAIN && "self-pipe read failed");
	mog_idleq_push(notifyrd->queue, mfd, MOG_QEV_RD);
}

static void notify_queue_step(struct mog_fd *mfd)
{
	switch (mfd->fd_type) {
	case MOG_FD_TYPE_NOTIFYRD: notifyrd_queue_step(mfd); return;
	case MOG_FD_TYPE_IOSTAT: mog_iostat_queue_step(mfd); return;
	default:
		assert(0 && mfd->fd_type && "bad fd_type in queue");
	}
}

/* this is the main loop of cmogstored */
void mog_notify_wait(bool need_usage_file)
{
	bool refresh;
	time_t next = usage_file_updated_at + usage_file_interval;
	time_t now = time(NULL);
	time_t timeout = next - now;
	struct mog_fd *mfd;

	refresh = __sync_bool_compare_and_swap(
	                   &notifywr->notes[MOG_NOTIFY_DEVICE_REFRESH], 1, 0);

	if (next <= now || refresh)
		global_mkusage();

	while ((mfd = mog_activeq_trytake(notifywr->queue)))
		notify_queue_step(mfd);

	if (need_usage_file == false)
		timeout = -1;
	else if (timeout > 0)
		timeout *= 1000;
	else
		timeout = 0;

	mog_intr_enable();
	mfd = mog_idleq_wait(notifywr->queue, timeout);
	mog_intr_disable();
	if (mfd)
		notify_queue_step(mfd);
}

/* this is async-signal safe */
void mog_notify(enum mog_notification note)
{
	ssize_t w;

	switch (note) {
	case MOG_NOTIFY_DEVICE_REFRESH:
		__sync_bool_compare_and_swap(&notifywr->notes[note], 0, 1);
		break;
	/* support changing thread counts later... */
	case MOG_NOTIFY_SIGNAL: break;
	default: assert(0 && "bad note passed");
	}

retry:
	w = write(notifywr_mfd->fd, "^", 1);
	if (w >= 0) return;

	switch (errno) {
	case_EAGAIN: return; /* somebody already woke this up */
	case EINTR: goto retry; /* just in case... */
	}

	/*
	 * we're screwed anyways, at least try this even though syslog()
	 * isn't safe inside a signal handler.
	 */
	syslog(LOG_CRIT, "mog_notify write() to pipe failed: %m");
	abort();
}
