/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
/*
 * a poll/select/libev/libevent-based implementation would have a hard time
 * migrating clients between threads
 */
#ifdef HAVE_KQUEUE
#include "activeq.h"
#include <sched.h>

struct mog_queue * mog_queue_new(void)
{
	int kqueue_fd = kqueue();

	if (kqueue_fd < 0)
		die("kqueue failed: %s\n", strerror(errno));

	return mog_queue_init(kqueue_fd);
}

static void check_cancel(void)
{
	mog_cancel_enable();
	pthread_testcancel();
	mog_cancel_disable();
}

static void intr_enable(int sig)
{
	sigset_t set;

	CHECK(int, 0, sigfillset(&set));
	CHECK(int, 0, sigdelset(&set, sig));
	CHECK(int, 0, pthread_sigmask(SIG_SETMASK, &set, NULL));
}

/*
 * grabs one active event off the event queue
 */
struct mog_fd * mog_idleq_wait(struct mog_queue *q, int timeout)
{
	int rc;
	struct mog_fd *mfd;
	struct kevent event;
	struct timespec ts;
	struct timespec *tsp;
	bool cancellable = timeout == MOG_CANCELLABLE;

	if (timeout < 0) {
		tsp = NULL;
	} else {
		ts.tv_sec = timeout / 1000;
		ts.tv_nsec = (timeout % 1000) * 1000000;
		tsp = &ts;
	}

retry:

	/*
	 * we enable SIGURG from pthread_kill() in thrpool.c when sleeping
	 * in kevent().  This allows us to wake up an respond to a
	 * cancellation request (since kevent() is not a cancellation point).
	 */
	if (cancellable) {
		check_cancel();
		intr_enable(SIGURG);
	}

	rc = kevent(q->queue_fd, NULL, 0, &event, 1, tsp);

	if (cancellable)
		PRESERVE_ERRNO( mog_intr_disable() );

	if (rc > 0) {
		mfd = event.udata;
		mog_fd_check_out(mfd);

		/* ignore pending cancel until the next round */
		return mfd;
	}
	if (cancellable)
		check_cancel();
	if (rc == 0) {
		if (timeout < 0)
			goto retry;
		return NULL;
	}

	if (errno == EINTR)
		goto retry;

	die("kevent(wait) failed with (%d): %s\n", rc, strerror(errno));
	return NULL;
}

MOG_NOINLINE static void
kevent_add_error(struct mog_queue *q, struct mog_fd *mfd)
{
	switch (errno) {
	case ENOMEM:
		syslog(LOG_ERR, "kevent(EV_ADD) out-of-space");
		if (mfd) {
			syslog(LOG_ERR, "falling back to active queue");
			mog_activeq_push(q, mfd);
		}
		return;
	default:
		syslog(LOG_ERR, "unhandled kevent(EV_ADD) error: %m");
		assert(0 && "BUG in our usage of kevent");
	}
}

static int add_event(int kqueue_fd, struct kevent *event)
{
	int rc;

	do {
		rc = kevent(kqueue_fd, event, 1, NULL, 0, NULL);
	} while (rc < 0 && errno == EINTR);

	return rc;
}

static void qpush(struct mog_queue *q, struct mog_fd *mfd, enum mog_qev ev)
{
	struct kevent event;
	u_short flags = EV_ADD | EV_ONESHOT;

	mfd->queue_state = MOG_QUEUE_STATE_OLD;
	EV_SET(&event, mfd->fd, (short)ev, flags, 0, 0, mfd);

	mog_fd_check_in(mfd);
	if (add_event(q->queue_fd, &event) != 0) {
		mog_fd_check_out(mfd);
		kevent_add_error(q, mfd);
	}
}

/*
 * Pushes in one mog_fd for kqueue to watch.
 *
 * Only call this from the mog_accept_loop *or*
 * if EAGAIN/EWOULDBLOCK is encountered in mog_queue_loop.
 */
void mog_idleq_push(struct mog_queue *q, struct mog_fd *mfd, enum mog_qev ev)
{
	if (ev == MOG_QEV_RW) {
		switch (mfd->fd_type) {
		case MOG_FD_TYPE_IOSTAT:
		case MOG_FD_TYPE_NOTIFYRD:
			ev = MOG_QEV_RD;
			break;
		case MOG_FD_TYPE_UNUSED:
		case MOG_FD_TYPE_ACCEPT:
		case MOG_FD_TYPE_FILE:
		case MOG_FD_TYPE_QUEUE:
		case MOG_FD_TYPE_SVC:
			assert(0 && "invalid fd_type for mog_idleq_push");
		default:
			ev = MOG_QEV_WR;
			break;
		}
	}
	qpush(q, mfd, ev);
}
#else /* ! HAVE_KQUEUE */
typedef int avoid_empty_file;
#endif /* ! HAVE_KQUEUE */
